@@ -2,7 +2,7 @@ package Perlbal::Plugin::PSGI;
 use strict;
 use warnings;
 use 5.008_001;
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 use Perlbal;
 use Plack::Util;
@@ -38,10 +38,13 @@ sub unregister {
 
 sub load {
     Perlbal::register_global_hook('manage_command.psgi_app', \&handle_psgi_app_command);
+    Perlbal::Service::add_role('psgi_server', sub { Perlbal::Plugin::PSGI::Client->new(@_) });
     return 1;
 }
 
 sub unload {
+    Perlbal::unregister_global_hook('manage_command.psgi_app');
+    Perlbal::Service::remove_role('psgi_server');
     return 1;
 }
 
@@ -101,6 +104,10 @@ sub close {
     # Do we need to do any cleanup?
 }
 
+sub forget_client {
+    # Do we need to do any cleanup?
+}
+
 sub write {
     my $self = shift;
     my $bufref = shift;
@@ -158,17 +165,27 @@ sub run_request {
     my $responder = sub {
         my $res = shift;
 
-        my $buf = "HTTP/1.0 $res->[0] @{[ HTTP::Status::status_message($res->[0]) ]}\015\012";
+        my $hd = $pb->{res_headers} = Perlbal::HTTPHeaders->new_response($res->[0]);
+        my %seen;
         while (my($k, $v) = splice @{$res->[1]}, 0, 2) {
-            $buf .="$k: $v\015\012";
+            if ($seen{lc($k)}++) {
+                my $newvalue = $hd->header($k) . "\015\012$k: $v";
+                $hd->header($k, $newvalue);
+            } else {
+                $hd->header($k, $v);
+            }
         }
-        $buf .= "\015\012";
-        $pb->write($buf);
+
+        $pb->setup_keepalive($hd);
+
+        $pb->state('xfer_resp');
+        $pb->tcp_cork(1);  # cork writes to self
+        $pb->write($hd->to_string_ref);
 
         if (!defined $res->[2]) {
             return Plack::Util::inline_object
                 write => sub { $pb->write(@_) },
-                close => sub { $pb->http_response_sent };
+                close => sub { $pb->write(sub { $pb->http_response_sent}) };
         } elsif (Plack::Util::is_real_fh($res->[2])) {
             $pb->reproxy_fh($res->[2], -s $res->[2]);
         } else {
@@ -191,9 +208,8 @@ Perlbal::Plugin::PSGI - PSGI web server on Perlbal
 
   LOAD PSGI
   CREATE SERVICE psgi
-    SET role    = web_server
+    SET role    = psgi_server
     SET listen  = 127.0.0.1:80
-    SET plugins = psgi
     PSGI_APP    = /path/to/app.psgi
   ENABLE psgi
 
@@ -202,9 +218,15 @@ Perlbal::Plugin::PSGI - PSGI web server on Perlbal
 This is a Perlbal plugin to allow any PSGI application run natively
 inside Perlbal process.
 
+=head1 COPYRIGHT
+
+Copyright 2009- Tatsuhiko Miyagawa
+
 =head1 AUTHOR
 
-Tatsuhiko Miyagawa E<lt>miyagawa@bulknews.netE<gt>
+Tatsuhiko Miyagawa
+
+Jonathan Steinert
 
 Based on Perlbal::Plugin::Cgilike written by Martin Atkins.
 
